import uuid
from datetime import datetime
from unittest.mock import patch, MagicMock, call

from django.test import TestCase, override_settings

from django_events_sourcing.register import disable_model_obj_dispatcher
from django_events_sourcing.testing import disable_dispatcher
from tests.test_app.models import Model1, ModelNoEvent, StatusModel, \
    StatusModel2


class TestModelDispatcher(TestCase):
    @patch('django_events_sourcing.dispatcher.dispatch')
    def setUp(self, mock) -> None:
        self.m1 = Model1.objects.create(
            int_field=10,
            char_field='test',
            uuid_field=uuid.uuid4(),
            dt_field=datetime(2019, 1, 1)
        )

    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_model_1_creation_no_event(self, mock):
        ModelNoEvent.objects.create(
            int_field=10,
            char_field='test',
            uuid_field=uuid.uuid4(),
            dt_field=datetime(2019, 1, 1)
        )
        mock.assert_not_called()

    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_model_1_creation_event_called(self, mock):
        m = Model1.objects.create(
            int_field=10,
            char_field='test',
            uuid_field=uuid.uuid4(),
            dt_field=datetime(2019, 1, 1)
        )
        mock.assert_called_once()
        mock.assert_called_with(
            'test_service',
            'model1__created',
            {
                'id': m.id,
                'int_field': 10,
                'char_field': 'test',
            },
            'amqp://guest:guest@localhost:5672/'
        )

    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_model_1_updated(self, mock):
        self.m1.int_field = 300
        self.m1.save()
        mock.assert_called_once()
        mock.assert_called_with(
            'test_service',
            'model1__updated',
            {
                'id': self.m1.id,
                'int_field': 300,
                'char_field': 'test',
            },
            'amqp://guest:guest@localhost:5672/'
        )

    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_model_1_deleted(self, mock):
        m1_id = self.m1.id
        self.m1.delete()
        mock.assert_called_once()
        mock.assert_called_with(
            'test_service',
            'model1__deleted',
            {
                'id': m1_id,
                'int_field': 10,
                'char_field': 'test',
            },
            'amqp://guest:guest@localhost:5672/'
        )

    @override_settings(MODELS_CRUD_EVENT=[])
    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_no_event_dispatched(self, mock):
        Model1.objects.create(
            int_field=10,
            char_field='test',
            uuid_field=uuid.uuid4(),
            dt_field=datetime(2019, 1, 1)
        )
        mock.assert_not_called()

    @override_settings(
        MODELS_CRUD_EVENT=[
            {'model': 'test_app.Model1', "event_name_prefix": "my_model1",
             'serializer': 'tests.test_app.serializers.Model1Serializer'}
        ]
    )
    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_custom_event_name_prefix(self, mock):
        m = Model1.objects.create(
            int_field=10,
            char_field='test',
            uuid_field=uuid.uuid4(),
            dt_field=datetime(2019, 1, 1)
        )
        mock.assert_called_once()
        mock.assert_called_with(
            'test_service',
            'my_model1__created',
            {
                'id': m.id,
                'int_field': 10,
                'char_field': 'test',
            },
            'amqp://guest:guest@localhost:5672/'
        )


class TestStatusModelDispatcher(TestCase):
    @patch('django_events_sourcing.dispatcher.dispatch')
    def setUp(self, mock) -> None:
        self.m1 = StatusModel.objects.create(
            int_field=10,
            char_field='test',
            uuid_field=uuid.uuid4(),
            dt_field=datetime(2019, 1, 1),
            status='created'
        )
        self.m2 = StatusModel2.objects.create(
            int_field=10,
            char_field='test',
            uuid_field=uuid.uuid4(),
            dt_field=datetime(2019, 1, 1),
            state='created'
        )

    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_status_dispatched(self, mock):
        m = StatusModel.objects.create(
            int_field=10,
            char_field='test',
            uuid_field=uuid.uuid4(),
            dt_field=datetime(2019, 1, 1),
            status='created'
        )
        mock.assert_called_with(
            'test_service',
            'status_model__created',
            {
                'id': m.id,
                'int_field': 10,
                'char_field': 'test',
                'uuid_field': str(m.uuid_field),
                'dt_field': '2019-01-01T00:00:00',
                'status': 'created'
            },
            'amqp://guest:guest@localhost:5672/'
        )

    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_different_status_on_creation(self, mock):
        m = StatusModel.objects.create(
            int_field=10,
            char_field='test',
            uuid_field=uuid.uuid4(),
            dt_field=datetime(2019, 1, 1),
            status='modified'
        )
        mock.assert_called_with(
            'test_service',
            'status_model__modified',
            {
                'id': m.id,
                'int_field': 10,
                'char_field': 'test',
                'uuid_field': str(m.uuid_field),
                'dt_field': '2019-01-01T00:00:00',
                'status': 'modified'
            },
            'amqp://guest:guest@localhost:5672/'
        )

    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_status_change_dispatched(self, mock):
        self.m1.status = 'failed'
        self.m1.int_field = 1000
        self.m1.save()
        mock.assert_called_with(
            'test_service',
            'status_model__failed',
            {
                'id': self.m1.id,
                'int_field': 1000,
                'char_field': 'test',
                'uuid_field': str(self.m1.uuid_field),
                'dt_field': '2019-01-01T00:00:00',
                'status': 'failed'
            },
            'amqp://guest:guest@localhost:5672/'
        )

    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_model_deleted_dispatched(self, mock):
        m1_id = self.m1.id
        self.m1.delete()
        mock.assert_called_with(
            'test_service',
            'status_model__deleted',
            {
                'id': m1_id,
                'int_field': 10,
                'char_field': 'test',
                'uuid_field': str(self.m1.uuid_field),
                'dt_field': '2019-01-01T00:00:00',
                'status': 'created'
            },
            'amqp://guest:guest@localhost:5672/'
        )

    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_model_different_state_field_name(self, mock):
        self.m2.state = 'failed'
        self.m2.int_field = 44
        self.m2.save()
        mock.assert_called_with(
            'test_service',
            'status_model2__failed',
            {
                'id': self.m2.id,
                'int_field': 44,
                'char_field': 'test',
                'uuid_field': str(self.m2.uuid_field),
                'dt_field': '2019-01-01T00:00:00',
                'state': 'failed'
            },
            'amqp://guest:guest@localhost:5672/'
        )


class TestDisableDispatcherDecorator(TestCase):
    @patch('django_events_sourcing.dispatcher.dispatch')
    def setUp(self, mock) -> None:
        self.m1 = Model1.objects.create(
            int_field=10,
            char_field='test',
            uuid_field=uuid.uuid4(),
            dt_field=datetime(2019, 1, 1)
        )
        self.m2 = Model1.objects.create(
            int_field=1,
            char_field='test2',
            uuid_field=uuid.uuid4(),
            dt_field=datetime(2020, 1, 1)
        )

    @disable_dispatcher()
    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_all_models_disabled(self, mock):
        self.m1.int_field = 100
        self.m1.save()
        self.m1.delete()
        Model1.objects.create(int_field=200, char_field='test',
                              uuid_field=uuid.uuid4(),
                              dt_field=datetime(2019, 1, 1))
        mock.assert_not_called()

    @disable_dispatcher(models_list=['test_app.Model1'])
    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_model_1_disabled(self, mock):
        self.m1.int_field = 100
        self.m1.save()
        self.m1.delete()
        StatusModel.objects.create(int_field=200, char_field='test',
                                   uuid_field=uuid.uuid4(),
                                   dt_field=datetime(2019, 1, 1))
        mock.assert_called_once()

    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_model_object_disabled(self, dispatcher):
        with disable_model_obj_dispatcher(self.m2):
            self.m1.int_field = 100
            self.m1.save()
            self.m2.int_field = 100
            self.m2.save()
        dispatcher.assert_called_once()
        dispatcher.assert_called_with(
            "test_service",
            "model1__updated",
            {
                'id': self.m1.id,
                'int_field': 100,
                'char_field': 'test',
            },
            'amqp://guest:guest@localhost:5672/'
        )

    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_model_obj_restored(self, dispatcher):
        with disable_model_obj_dispatcher(self.m2):
            self.m1.int_field = 100
            self.m1.save()
            self.m2.int_field = 100
            self.m2.save()
        self.m2.int_field = 0
        self.m2.save()
        dispatcher.assert_has_calls(
            [
                call(
                    "test_service", "model1__updated",
                    {'id': self.m1.id, 'int_field': 100, 'char_field': 'test'},
                    'amqp://guest:guest@localhost:5672/'
                ),
                call(
                    "test_service", "model1__updated",
                    {'id': self.m2.id, 'int_field': 0, 'char_field': 'test2'},
                    'amqp://guest:guest@localhost:5672/'
                ),
            ]
        )


@disable_dispatcher()
class TestDisableDispatcherClassDecoratorAllModels(TestCase):
    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_nothing_dispatched(self, mock):
        StatusModel.objects.create(int_field=200, char_field='test',
                                   uuid_field=uuid.uuid4(),
                                   dt_field=datetime(2019, 1, 1))
        mock.assert_not_called()


@disable_dispatcher(models_list=['test_app.Model1'])
class TestDisableDispatcherClassDecoratorModel1(TestCase):
    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_model_1_not_dispatched(self, mock):
        self.m1 = Model1.objects.create(
            int_field=10,
            char_field='test',
            uuid_field=uuid.uuid4(),
            dt_field=datetime(2019, 1, 1)
        )
        mock.assert_not_called()

    @patch('django_events_sourcing.dispatcher.dispatch')
    def test_status_model_dispatched(self, mock):
        StatusModel.objects.create(int_field=200, char_field='test',
                                   uuid_field=uuid.uuid4(),
                                   dt_field=datetime(2019, 1, 1))
        mock.assert_called_once()



from django.test.utils import TestContextDecorator

from django_events_sourcing.register import unregister_models, register_models


class disable_dispatcher(TestContextDecorator):
    """
    Decorator that disables any dispatching that would occur using
    django_events_sourcing.
    It can also be used as a context manager using "with" statement.
    """
    def __init__(self, models_list=None, **kwargs):
        self.models_list = models_list
        super().__init__()

    def enable(self):
        unregister_models(self.models_list)

    def disable(self):
        register_models(self.models_list)



from django.apps import apps
from django.conf import settings


def load_model_configurations(sender):
    """
    Iterates through setting MODELS_CRUD_EVENT searching for the sender
    model configurations.
    :param sender: Django Model
    :return dict
    """
    for model_config in settings.MODELS_CRUD_EVENT:
        model = model_config['model']
        app, model = model.rsplit('.', 1)
        model = apps.get_app_config(app).get_model(model)
        if sender == model:
            return model_config
    return None # Not found.



from importlib import import_module

import re

from django.conf import settings
from rest_framework import serializers

from django_events_sourcing.utils import load_model_configurations
from django_events_sourcing.senders.kombu import dispatch


def slug_model_name(instance):
    return re.sub('(?<!^)(?=[A-Z])', '_', type(instance).__name__).lower()


def get_event_name(instance, model_data, action):
    event_name = model_data.get("event_name_prefix")
    if event_name is None:
        event_name = slug_model_name(instance)
    if action == 'deleted':
        event_name += '__deleted'
    elif not model_data.get('status_field') and action == 'created':
        event_name += '__created'
    elif not model_data.get('status_field'):
        event_name += '__updated'
    else:
        event_name += '__{}'.format(
            getattr(instance, model_data['status_field']))
    return event_name


def serialize_model(instance, model_data):
    serializer_cls = model_data.get('serializer')
    if serializer_cls:
        serializer_module, serializer = serializer_cls.rsplit('.', 1)
        serializer = getattr(import_module(serializer_module), serializer)
    else:
        serializer = type('NoClassSerializer', (serializers.ModelSerializer,),
                          {'Meta': type('Meta', (), {'model': type(instance),
                                                     'fields': '__all__'})})
    serializer = serializer(instance=instance)
    return serializer.data


def dispatch_event(instance, action):
    model_data = load_model_configurations(type(instance))
    if model_data is None:
        return
    event_name = get_event_name(instance, model_data, action)
    serialized_model = serialize_model(instance, model_data)

    try:
        amqp_uri = settings.AMQP_URI
    except AttributeError:
        # Fall back to old method of getting it
        amqp_uri = settings.NAMEKO_CONFIG.get("AMQP_URI")

    dispatch(
        settings.SERVICE_NAME, event_name, serialized_model, amqp_uri
    )



from django.db import models

# Create your models here.
class ModelNoEvent(models.Model):
    int_field = models.IntegerField()
    char_field = models.CharField(max_length=100)
    uuid_field = models.UUIDField()
    dt_field = models.DateTimeField()


class Model1(models.Model):
    int_field = models.IntegerField()
    char_field = models.CharField(max_length=100)
    uuid_field = models.UUIDField()
    dt_field = models.DateTimeField()


class StatusModel(models.Model):
    STATUS_CHOICES = (('created', 'Created'),
                      ('modified', 'Modified'),
                      ('finished', 'Finished'),
                      ('failed', 'Failed'))

    status = models.CharField(max_length=100, choices=STATUS_CHOICES)
    int_field = models.IntegerField()
    char_field = models.CharField(max_length=100)
    uuid_field = models.UUIDField()
    dt_field = models.DateTimeField()


class StatusModel2(models.Model):
    STATUS_CHOICES = (('created', 'Created'),
                      ('modified', 'Modified'),
                      ('finished', 'Finished'),
                      ('failed', 'Failed'))

    state = models.CharField(max_length=100, choices=STATUS_CHOICES)
    int_field = models.IntegerField()
    char_field = models.CharField(max_length=100)
    uuid_field = models.UUIDField()
    dt_field = models.DateTimeField()


from rest_framework import serializers

from tests.test_app.models import Model1


class Model1Serializer(serializers.ModelSerializer):
    class Meta:
        model = Model1
        fields = ('id', 'int_field', 'char_field', )



from django.apps import AppConfig


class DjangoEventsSourcingConfig(AppConfig):
    name = 'django_events_sourcing'

    def ready(self):
        from django_events_sourcing.register import register_models
        register_models()



import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SECRET_KEY = '9%%8-rz9+nu+&%0=nf#64g!j2m^=m^r=)(687tvjqq9jm^(ok5'
INSTALLED_APPS = (
    'django.contrib.contenttypes',
    'django.contrib.staticfiles',
    'django.contrib.auth',
    'rest_framework',
    'django_events_sourcing',
    'tests.test_app',
)


DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}


SERVICE_NAME = 'test_service'
AMQP_URI = 'amqp://guest:guest@localhost:5672/'

MODELS_CRUD_EVENT = [
    {'model': 'test_app.Model1',
     'serializer': 'tests.test_app.serializers.Model1Serializer'},
    {'model': 'test_app.StatusModel', 'status_field': 'status'},
    {'model': 'test_app.StatusModel2', 'status_field': 'state'},
]



from contextlib import contextmanager

import amqp
from kombu import Exchange, Connection, producers

DEFAULT_HEARTBEAT = 60
DEFAULT_TRANSPORT_OPTIONS = {
    'max_retries': 3,
    'interval_start': 2,
    'interval_step': 1,
    'interval_max': 5
}
DEFAULT_RETRY_POLICY = {'max_retries': 3}

default_transport_options = {
    'confirm_publish': True
}


def get_connection(amqp_uri, heartbeat=None, ssl=None, transport_options=None):
    if heartbeat is None: heartbeat = DEFAULT_HEARTBEAT
    if transport_options is None: transport_options = default_transport_options.copy()

    connection = Connection(
        amqp_uri, transport_options=transport_options,
        heartbeat=heartbeat, ssl=ssl
    )
    return connection


@contextmanager
def get_producer(
        amqp_uri, confirms=True, ssl=None, transport_options=None,
        heartbeat=None
):
    if transport_options is None:
        transport_options = DEFAULT_TRANSPORT_OPTIONS.copy()

    transport_options['confirm_publish'] = confirms
    conn = get_connection(
        amqp_uri, heartbeat=heartbeat, ssl=ssl,
        transport_options=transport_options
    )
    with producers[conn].acquire(block=True) as producer:
        yield producer


def get_exchange(service_name):
    exchange_name = "%s.events" % service_name
    return Exchange(exchange_name, type='topic', auto_delete=True)



def dispatch(
        service_name, event_name, payload, amqp_uri, ssl=None,
        transport_options=None, heartbeat=60,
):
    """
    Dispatches an event. It mimics Nameko's and Micro-Framework's format of
    receiving events.

    :param event_name: Name of the Event
    :param payload: Serializable Payload to be sent
    :param amqp_uri: URI to connect to AMQP
    :param service_name: Name of the Service sending the event
    :param ssl: To use SSL or not
    :param transport_options: Transport Options dict
    :param heartbeat: AMQP Heartbeats interval
    """
    try:
        exchange = get_exchange(service_name)
        with get_producer(
                amqp_uri,
                ssl=ssl,
                transport_options=transport_options,
                heartbeat=heartbeat
        ) as producer:
            producer.publish(
                payload, routing_key=event_name, exchange=exchange
            )
    except amqp.exceptions.NotFound:
        pass
